@@ -1,11 +1,11 @@
 from .gpr_iterative import GaussianProcess
 from .gpr_batch import GaussianProcess
-from .gpr_batch_internal import GaussianProcessInternal
+# from .gpr_batch_internal import GaussianProcessInternal
 from .pytorch_kernel import FPKernel, FPKernelNoforces
-from .pytorch_kernel_internal import FPKernelInternal
+# from .pytorch_kernel_internal import FPKernelInternal
 from .pytorch_kerneltypes import SquaredExp
 from .calculator import GPRCalculator
 from .prior import ConstantPrior
 
 __all__ = ["gpr_iterative", "gpr_batch", "gpr_batch_internal", "pytorch_kernel", "pytorch_kernel_internal", "pytorch_kerneltypes", "calculator", "prior",
-           "GaussianProcess", "GaussianProcessInternal", "FPKernel", "FPKernelNoforces", "FPKernelInternal", "SquaredExp", "GPRCalculator", "ConstantPrior"]
+           "GaussianProcess", "FPKernel", "FPKernelNoforces", "SquaredExp", "GPRCalculator", "ConstantPrior"]
