@@ -9,9 +9,9 @@ from ase.optimize import FIRE, MDMin, LBFGS, BFGS
 from ase.parallel import parprint, parallel_function
 
 try:
-    from ase.mep import NEB
+    from ase.mep import NEB, DyNEB
 except ModuleNotFoundError:
-    from ase.neb import NEB
+    from ase.neb import NEB, DyNEB
 
 from aenet_gpr.src import GPRCalculator
 from aenet_gpr.util import ReferenceData
@@ -33,7 +33,7 @@ def min_cartesian_dist(img, train_images):
 class AIDNEB:
 
     def __init__(self, start, end, input_param: InputParameters, model_calculator=None, calculator=None,
-                 interpolation='idpp', n_images=15, k=None, mic=False,
+                 interpolation='idpp', n_images=15, n_train_images=3, k=None, mic=False,
                  neb_method='improvedtangent',  # 'improvedtangent', 'aseneb'
                  remove_rotation_and_translation=False,
                  max_train_data=25, force_consistent=None,
@@ -191,6 +191,7 @@ class AIDNEB:
         self.start = start
         self.end = end
         self.n_images = n_images
+        self.n_train_images = n_train_images
         self.mic = mic
         self.rrt = remove_rotation_and_translation
         self.neb_method = neb_method
@@ -221,45 +222,46 @@ class AIDNEB:
             for i in range(10000):
                 mic_images += [mic_initial.copy()]
             mic_images += [mic_final]
-            neb_mic = NEB(mic_images, climb=False, method=self.neb_method,
-                          remove_rotation_and_translation=self.rrt)
+            neb_mic = DyNEB(mic_images, climb=False, method=self.neb_method, remove_rotation_and_translation=self.rrt)
             neb_mic.interpolate(method='linear', mic=self.mic)
             self.i_endpoint.positions = mic_images[1].positions[:]
             self.e_endpoint.positions = mic_images[-2].positions[:]
 
         # Calculate the initial and final end-points (if necessary).
-        self.i_endpoint.calc = copy.deepcopy(self.ase_calc)
-        self.e_endpoint.calc = copy.deepcopy(self.ase_calc)
+        if self.i_endpoint.calc is None:
+            self.i_endpoint.calc = copy.deepcopy(self.ase_calc)
+        if self.e_endpoint.calc is None:
+            self.e_endpoint.calc = copy.deepcopy(self.ase_calc)
         self.i_endpoint.get_potential_energy(force_consistent=force_consistent)
         self.i_endpoint.get_forces()
         self.e_endpoint.get_potential_energy(force_consistent=force_consistent)
         self.e_endpoint.get_forces()
 
+        if isinstance(self.i_endpoint, Atoms):
+            io.write('initial.traj', self.i_endpoint)
+        if isinstance(self.e_endpoint, Atoms):
+            io.write('final.traj', self.e_endpoint)
+
         # Calculate the distance between the initial and final endpoints.
-        d_start_end = np.sum((self.i_endpoint.positions.flatten() -
-                              self.e_endpoint.positions.flatten()) ** 2) ** 0.5
-
-        self.rmax = 0.2 * d_start_end
+        self.d_start_end = np.sum((self.i_endpoint.positions.flatten() -
+                                   self.e_endpoint.positions.flatten()) ** 2) ** 0.5
 
+        # self.rmax = 0.4 * self.d_start_end
         # A) Create images using interpolation if user does define a path.
         if interp_path is None:
             if isinstance(self.n_images, float):
-                self.n_images = int(d_start_end / self.n_images)
+                self.n_images = int(self.d_start_end / self.n_images)
             if self.n_images <= 3:
                 self.n_images = 3
             self.images = make_neb(self)
-            raw_spring = 1. * np.sqrt(self.n_images - 1) / d_start_end  # 1 or 2?
-            self.spring = np.clip(raw_spring, 0.05, 0.1)
+            if self.spring is None:
+                self.spring = 0.1 * (self.n_images - 1) / self.d_start_end
+                # self.spring = 2. * np.sqrt(self.n_images - 1) / self.d_start_end ** 2
 
-            neb_interpolation = NEB(self.images, climb=False, k=self.spring,
-                                    method=self.neb_method,
-                                    remove_rotation_and_translation=self.rrt)
+            neb_interpolation = DyNEB(self.images, climb=False, k=self.spring, method=self.neb_method, remove_rotation_and_translation=self.rrt)
             neb_interpolation.interpolate(method='linear', mic=self.mic)
             if interpolation == 'idpp':
-                neb_interpolation = NEB(
-                    self.images, climb=True,
-                    k=self.spring, method=self.neb_method,
-                    remove_rotation_and_translation=self.rrt)
+                neb_interpolation = DyNEB(self.images, climb=True, k=self.spring, method=self.neb_method, remove_rotation_and_translation=self.rrt)
                 neb_interpolation.interpolate(method='idpp', mic=self.mic)
                 # neb_interpolation.idpp_interpolate(optimizer=FIRE, mic=self.mic)
 
@@ -282,16 +284,25 @@ class AIDNEB:
 
         # Guess spring constant (k) if not defined by the user.
         if self.spring is None:
-            raw_spring = 1. * np.sqrt(self.n_images - 1) / d_start_end  # 1 or 2?
-            self.spring = np.clip(raw_spring, 0.05, 0.10)
+            self.spring = 0.1 * (self.n_images - 1) / self.d_start_end
+            # self.spring = 2. * np.sqrt(self.n_images - 1) / self.d_start_end ** 2
         # Save initial interpolation.
         self.initial_interpolation = self.images[:]
 
-        print('d_start_end: ', d_start_end)
-        print('r_max: ', self.rmax)
+        print()
+        print('Distance between initial and final: ', self.d_start_end)
+        # print(f"r_max (threshold to prevent over-relaxation when training data is sparse): {self.rmax:.4f}")
         print('spring_constant: ', self.spring)
 
-    def run(self, fmax=0.05, unc_convergence=0.05, dt=0.05, ml_steps=150, optimizer="FIRE", max_unc_trheshold=1.0):
+    def run(self,
+            fmax=0.05,
+            unc_convergence=0.02,
+            dt=0.1,
+            ml_steps=150,
+            optimizer="MDMin",
+            update_step=10,
+            check_ref_force=False,
+            climbing=True):
 
         """
         Executing run will start the NEB optimization process.
@@ -341,120 +352,95 @@ class AIDNEB:
                          filename=trajectory_observations,
                          restart=self.use_previous_observations)
 
+        print(f"[INFO] Number of initial training data: {self.n_train_images} including initial and final")
+        n_to_add = min(self.n_images - 2, max(1, int(self.d_start_end / 3.0)))
+        if (n_to_add + 2) > self.n_train_images:
+            print(f"[INFO] Distance between initial and final (ΔR = {self.d_start_end} Å) is too large")
+            print(f"[INFO] Use {n_to_add + 2} initial training data instead of {self.n_train_images}")
+            self.n_train_images = n_to_add + 2
+
         train_images = io.read(trajectory_observations, ':')
         if len(train_images) == 2:
-            middle = int(self.n_images * (2. / 3.))
-            e_is = self.i_endpoint.get_potential_energy()
-            e_fs = self.e_endpoint.get_potential_energy()
+            # middle = int(self.n_images * (2. / 3.))
+            # e_is = self.i_endpoint.get_potential_energy()
+            # e_fs = self.e_endpoint.get_potential_energy()
+            #
+            # if e_is > e_fs:
+            #     middle = int(self.n_images * (1. / 3.))
+            n_to_add = self.n_train_images - 2
+            middle_indices = [int(round(i * (self.n_images - 1) / (n_to_add + 1))) for i in range(1, n_to_add + 1)]
 
-            if e_is > e_fs:
-                middle = int(self.n_images * (1. / 3.))
+            for middle_idx in middle_indices:
+                print(f"[INFO] Adding training image at NEB index {middle_idx}")
 
-            self.atoms.positions = self.images[middle].get_positions()
-            self.atoms.calc = self.ase_calc
-            self.atoms.get_potential_energy(force_consistent=self.force_consistent)
-            self.atoms.get_forces()
-            dump_observation(atoms=self.atoms, method='neb',
-                             filename=trajectory_observations,
-                             restart=self.use_previous_observations)
-            self.function_calls += 1
-            self.force_calls += 1
-            self.step += 1
+                self.atoms.positions = self.images[middle_idx].get_positions()
+                self.atoms.calc = self.ase_calc
+
+                # Reference calculation
+                self.atoms.get_potential_energy(force_consistent=self.force_consistent)
+                self.atoms.get_forces()
+                dump_observation(atoms=self.atoms, method='neb',
+                                 filename=trajectory_observations,
+                                 restart=self.use_previous_observations)
+                self.function_calls += 1
+                self.force_calls += 1
+
+        else:
+            self.function_calls = len(train_images)
+            self.force_calls = len(train_images)
+
+        self.step += 1
 
         weight_update = self.input_param.weight
         scale_update = self.input_param.scale
 
         self.rmin = 0.1
-        max_weight = 5.0
-        violated_index = 0
-        set_update_step = False
-        update_step = 10
+        self.max_unc_hist = []
 
         while True:
 
             # 0. Start from initial interpolation every 50 steps.
-            if self.step % 50 == 0:
-                parprint('Starting from initial interpolation...')
-                self.images = copy.deepcopy(self.initial_interpolation)
+            # if self.step % 50 == 0:
+            #     parprint('Starting from initial interpolation...')
+            #     self.images = copy.deepcopy(self.initial_interpolation)
 
             # 1. Collect observations.
-            # This serves to use_previous_observations from a previous
-            # (and/or parallel) runs.
+            # This serves to use_previous_observations from a previous (and/or parallel) runs.
             train_images = io.read(trajectory_observations, ':')
 
             train_data = ReferenceData(structure_files=train_images,
                                        file_format='ase',
                                        device=self.input_param.device,
                                        descriptor=self.input_param.descriptor,
+                                       standardization=self.input_param.standardization,
                                        data_type=self.input_param.data_type,
                                        data_process=self.input_param.data_process,
                                        soap_param=self.input_param.soap_param,
-                                       standardization=False,
+                                       mace_param=self.input_param.mace_param,
                                        mask_constraints=self.input_param.mask_constraints)
 
             if train_data.standardization:
                 train_data.standardize_energy_force(train_data.energy)
 
+            if check_ref_force:
+                self.input_param.filter = False
+
             # 2. Prepare a calculator.
             print('Training data size: ', len(train_images))
             print('Descriptor: ', self.input_param.descriptor)
 
-            if violated_index is None and not set_update_step:
-                update_step = self.step
-                set_update_step = True
-
-            if self.step % 25 == update_step:
+            if update_step is not None and self.step >= update_step:
+                update_step *= 2
                 self.input_param.fit_weight = True
                 self.input_param.fit_scale = True
 
-                while True:
-                    if self.input_param.filter:
-                        train_data.filter_similar_data(threshold=self.rmin)
-                        print('Actual training data size (after removing similar data): ', len(train_data.images))
-
-                    try:
-                        train_data.config_calculator(kerneltype='sqexp',
-                                                     scale=scale_update,
-                                                     weight=weight_update,
-                                                     noise=self.input_param.noise,
-                                                     noisefactor=self.input_param.noisefactor,
-                                                     use_forces=self.input_param.use_forces,
-                                                     sparse=self.input_param.sparse,
-                                                     sparse_derivative=self.input_param.sparse_derivative,
-                                                     autograd=self.input_param.autograd,
-                                                     train_batch_size=self.input_param.train_batch_size,
-                                                     eval_batch_size=self.input_param.eval_batch_size,
-                                                     fit_weight=self.input_param.fit_weight,
-                                                     fit_scale=self.input_param.fit_scale)
-
-                        if train_data.calculator.weight < max_weight:
-                            break
-                        else:
-                            raise ValueError(f"Weight parameter too high ({train_data.calculator.weight}).")
-
-                    except Exception as e:
-                        print(f"{e} Increase r_min to {self.rmin + 0.1} and retry.")
-                        self.rmin += 0.1
-
-                        train_data = ReferenceData(structure_files=train_images,
-                                                   file_format='ase',
-                                                   device=self.input_param.device,
-                                                   descriptor=self.input_param.descriptor,
-                                                   data_type=self.input_param.data_type,
-                                                   data_process=self.input_param.data_process,
-                                                   soap_param=self.input_param.soap_param,
-                                                   standardization=False,
-                                                   mask_constraints=self.input_param.mask_constraints)
-
-            else:
-                self.input_param.fit_weight = False
-                self.input_param.fit_scale = False
-
                 if self.input_param.filter:
                     train_data.filter_similar_data(threshold=self.rmin)
                     print('Actual training data size (after removing similar data): ', len(train_data.images))
 
-                train_data.config_calculator(kerneltype='sqexp',
+                train_data.config_calculator(prior=self.input_param.prior,
+                                             prior_update=self.input_param.prior_update,
+                                             kerneltype='sqexp',
                                              scale=scale_update,
                                              weight=weight_update,
                                              noise=self.input_param.noise,
@@ -468,8 +454,31 @@ class AIDNEB:
                                              fit_weight=self.input_param.fit_weight,
                                              fit_scale=self.input_param.fit_scale)
 
-            # self.rmin = 0.1
-            print('r_min: ', self.rmin)
+            else:
+                self.input_param.fit_weight = False
+                self.input_param.fit_scale = False
+
+                if self.input_param.filter:
+                    train_data.filter_similar_data(threshold=self.rmin)
+                    print('Actual training data size (after removing similar data): ', len(train_data.images))
+
+                train_data.config_calculator(prior=self.input_param.prior,
+                                             prior_update=self.input_param.prior_update,
+                                             kerneltype='sqexp',
+                                             scale=scale_update,
+                                             weight=weight_update,
+                                             noise=self.input_param.noise,
+                                             noisefactor=self.input_param.noisefactor,
+                                             use_forces=self.input_param.use_forces,
+                                             sparse=self.input_param.sparse,
+                                             sparse_derivative=self.input_param.sparse_derivative,
+                                             autograd=self.input_param.autograd,
+                                             train_batch_size=self.input_param.train_batch_size,
+                                             eval_batch_size=self.input_param.eval_batch_size,
+                                             fit_weight=self.input_param.fit_weight,
+                                             fit_scale=self.input_param.fit_scale)
+
+            print(f"r_min (threshold to avoid adding nearly-duplicate training data): {self.rmin:.2f}")
             print('GPR model hyperparameters: ', train_data.calculator.hyper_params)
 
             self.model_calculator = GPRCalculator(calculator=train_data.calculator, train_data=train_data)
@@ -477,85 +486,61 @@ class AIDNEB:
             scale_update = train_data.calculator.scale.clone().detach().item()
 
             # Detach calculator from the prev. optimized images (speed up).
-            # for i in self.images:
-            #     i.calc = None
-            # # Train only one process.
+            for i in self.images:
+                i.calc = None
+            # Train only one process.
             # calc.update_train_data(train_images, test_images=self.images)
-            # Attach the calculator (already trained) to each image.
+            # Attach the trained calculator to each image.
             for i in self.images:
                 i.calc = copy.deepcopy(self.model_calculator)
 
             # 3. Optimize the NEB in the predicted PES.
-            # Get path uncertainty for deciding whether NEB or CI-NEB.
-            predictions = get_neb_predictions(self.images)
-            neb_pred_uncertainty = predictions['uncertainty']
-
-            # Climbing image NEB mode is risky when the model is trained
-            # with a few data points. Switch on climbing image (CI-NEB) only
-            # when the uncertainty of the NEB is low.
+            # Use previous path uncertainty for deciding whether NEB or CI-NEB.
+            # Climbing image NEB mode is risky when the model is trained with a few data points.
+            # Switch on climbing image only when the uncertainty of the NEB the force of the climbing image are low.
             climbing_neb = False
-            if np.max(neb_pred_uncertainty) <= unc_convergence:
-                parprint('Climbing image is now activated.')
-                climbing_neb = True
+            if climbing:
+                if self.step > 1 and self.max_unc_hist[-1] <= unc_convergence:
+                    parprint(f"Climbing image is now activated.")
+                    climbing_neb = True
+            else:
+                pass
 
-            ml_neb = NEB(self.images, climb=climbing_neb, method=self.neb_method, k=self.spring)
+            ml_neb = DyNEB(self.images, climb=climbing_neb, method=self.neb_method, k=self.spring)
             # FIRE, MDMin, LBFGS, BFGS
             if optimizer.lower() == 'mdmin':
-                neb_opt = MDMin(ml_neb, dt=dt, trajectory=self.trajectory)
+                neb_opt = MDMin(ml_neb, dt=dt, trajectory="gpr_neb.traj")
             elif optimizer.lower() == 'lbfgs':
-                neb_opt = LBFGS(ml_neb, trajectory=self.trajectory)
+                neb_opt = LBFGS(ml_neb, trajectory="gpr_neb.traj")
             elif optimizer.lower() == 'bfgs':
-                neb_opt = BFGS(ml_neb, trajectory=self.trajectory)
+                neb_opt = BFGS(ml_neb, trajectory="gpr_neb.traj")
             else:
-                neb_opt = FIRE(ml_neb, dt=dt, trajectory=self.trajectory)
+                neb_opt = FIRE(ml_neb, dt=dt, trajectory="gpr_neb.traj")
 
-            # Safe check to optimize the images.
-            if np.max(neb_pred_uncertainty) <= max_unc_trheshold:
-                # neb_opt.run(fmax=(fmax * 0.8), steps=ml_steps)
+            # Optimize the images
+            neb_opt.run(fmax=fmax * 1.0, steps=ml_steps)
 
-                # 직전 스텝의 전체 밴드 좌표 스냅샷(롤백용)
-                prev_positions = [im.get_positions().copy() for im in self.images]
+            nim = len(self.images) - 2
+            nat = len(self.images[0])
 
-                for step in range(ml_steps):
-                    # 한 스텝만 진행
-                    neb_opt.run(fmax=fmax * 0.8, steps=1)
+            F = ml_neb.get_forces()  # (n_mobile*nat*3) flat
+            F = F.reshape(nim, nat, 3)
+            max_f_image = np.sqrt((F ** 2).sum(-1)).max().item()
 
-                    # 신뢰반경 검사(엔드포인트 제외: 1..len(images)-2)
-                    violated_index = None
-                    for i in range(1, len(self.images) - 1):
-                        d2 = min_cartesian_dist(self.images[i], train_images)
-                        if d2 > (self.rmax ** 2):
-                            violated_index = i
-                            print("Relaxation early stop: the distance from current image {0} to the nearest training data is larger than r_max".format(violated_index))
-                            break
-
-                    if violated_index is not None:
-                        for im, pos in zip(self.images, prev_positions):
-                            im.set_positions(pos)  # 마지막 GPR 스텝 취소
-
-                        break
-                    else:
-                        # 스텝 확정: 다음 롤백 기준 업데이트
-                        prev_positions = [im.get_positions().copy() for im in self.images]
-
-                    # (선택) 수렴하면 조기 종료
-                    if neb_opt.converged():
-                        break
-
-            else:
-                print("The uncertainty of the NEB lies above the max_unc threshold (1.0).")
-                print("NEB won't be optimized and the image with maximum uncertainty is just evaluated and added")
-
-            predictions = get_neb_predictions(self.images)
+            predictions = get_neb_predictions(self.images[1:-1])
             neb_pred_energy = predictions['energy']
             neb_pred_uncertainty = predictions['uncertainty']
 
             # 5. Print output.
             max_e = np.max(neb_pred_energy)
-            max_e_ind = np.argsort(neb_pred_energy)[-1]
-            # train_images 의 calc 는 reference,
-            # self.images 의 calc 는 GP
-            max_f = get_fmax(self.images[max_e_ind])  # get_fmax(train_images[-1])
+            max_unc = np.max(neb_pred_uncertainty)
+            self.max_unc_hist.append(max_unc)
+
+            # Calculator of train_images is reference, while Calculator of self.images is GP
+            if check_ref_force:
+                max_f = get_fmax(train_images[-1])
+            else:
+                max_f = max_f_image
 
             pbf = max_e - self.i_endpoint.get_potential_energy(force_consistent=self.force_consistent)
             pbb = max_e - self.e_endpoint.get_potential_energy(force_consistent=self.force_consistent)
@@ -567,19 +552,20 @@ class AIDNEB:
             parprint('Predicted barrier (-->):', pbf)
             parprint('Predicted barrier (<--):', pbb)
             parprint('Number of images:', len(self.images))
-            parprint('Max. uncertainty:', np.max(neb_pred_uncertainty))
-            parprint("Max. force:", max_f.item())
+            parprint('Max. uncertainty:', max_unc)
+            parprint("Max. force:", max_f)
             msg = "--------------------------------------------------------\n"
             parprint(msg)
 
             # 6. Check convergence.
+            ok_unc = max_unc <= unc_convergence
+
             # Max.forces and NEB images uncertainty must be below *fmax* and *unc_convergence* thresholds.
-            if len(train_images) > 2 and max_f <= fmax and np.max(neb_pred_uncertainty[1:-1]) <= unc_convergence:
+            if len(train_images) > 2 and max_f <= fmax and ok_unc and (climbing_neb or not climbing):
                 parprint('A saddle point was found.')
 
-                # if np.max(neb_pred_uncertainty[1:-1]) < unc_convergence:
                 io.write(self.trajectory, self.images)
-                parprint('Uncertainty of the images above threshold.')
+                parprint('Uncertainty of the images below threshold.')
                 parprint('NEB converged.')
                 parprint('The NEB path can be found in:', self.trajectory)
                 msg = "Visualize the last path using 'ase gui "
@@ -591,54 +577,62 @@ class AIDNEB:
             # Candidates are the optimized NEB images in the predicted PES.
             candidates = copy.deepcopy(self.images)[1:-1]
 
-            if violated_index is not None:
-                best_candidate = candidates[violated_index]
-                sorted_candidates = False
+            if max_unc > unc_convergence:
+                sorted_candidates = acquisition(train_images=train_images,
+                                                candidates=candidates,
+                                                mode='uncertainty',
+                                                objective='max')
             else:
-                if np.max(neb_pred_uncertainty) > unc_convergence:
+                if self.step % 5 == 0:
                     sorted_candidates = acquisition(train_images=train_images,
                                                     candidates=candidates,
-                                                    mode='uncertainty',
-                                                    objective='max')
+                                                    mode='fmax',
+                                                    objective='min')
                 else:
-                    if self.step % 5 == 0:
-                        sorted_candidates = acquisition(train_images=train_images,
-                                                        candidates=candidates,
-                                                        mode='fmax',
-                                                        objective='min')
-                    else:
-                        sorted_candidates = acquisition(train_images=train_images,
-                                                        candidates=candidates,
-                                                        mode='ucb',
-                                                        objective='max')
+                    sorted_candidates = acquisition(train_images=train_images,
+                                                    candidates=candidates,
+                                                    mode='ucb',
+                                                    objective='max')
 
             # Select the best candidate.
-            accepted = False
+            if check_ref_force or not self.input_param.filter:
+                chosen_candidate = sorted_candidates.pop(0)
+            else:
+                accepted = False
+                fp_train = train_data.generate_cartesian(train_data.images)
+                N = fp_train.shape[0]
 
-            fp_train = train_data.generate_cartesian(train_data.images)
-            N = fp_train.shape[0]
+                while not accepted:
+                    sorted_candidates_tmp = copy.deepcopy(sorted_candidates)
+                    found_candidate = False
+    
+                    while sorted_candidates_tmp:
+                        best_candidate = sorted_candidates_tmp.pop(0)
+                        fp_candidate = train_data.generate_cartesian_per_data(best_candidate).flatten()
 
-            while sorted_candidates and not accepted:
-                best_candidate = sorted_candidates.pop(0)
-                fp_candidate = train_data.generate_cartesian_per_data(best_candidate).flatten()
+                        for i in range(N):
+                            xi = fp_train[i].flatten()
+                            dist = torch.linalg.norm(xi - fp_candidate)
 
-                for i in range(N):
-                    xi = fp_train[i].flatten()
-                    dist = torch.linalg.norm(xi - fp_candidate)
+                            if dist < self.rmin:
+                                print(f"Candidate rejected: too close to train data {i} (dist={dist:.4f} < r_min={self.rmin:.3f})")
+                                break
+                        else:
+                            accepted = True
+                            found_candidate = True
+                            chosen_candidate = best_candidate
+                            break
 
-                    if dist < self.rmin:
-                        print(f"Candidate is reclined to be train data since distance with previous train data {i}: {dist:.4f} < r_min {self.rmin}")
-                        break
-                else:
-                    accepted = True
-                    # print("Candidate accepted")
+                    if not found_candidate and not accepted:
+                        self.rmin -= 0.01
+                        print(f"No candidate accepted. Decrease r_min to {self.rmin:.3f} and retry...")
 
             # Save the other candidates for multi-task optimization.
             if sorted_candidates:
                 io.write(trajectory_candidates, sorted_candidates)
 
             # 8. Evaluate the target function and save it in *observations*.
-            self.atoms.positions = best_candidate.get_positions()
+            self.atoms.positions = chosen_candidate.get_positions()
             self.atoms.calc = self.ase_calc
             self.atoms.get_potential_energy(force_consistent=self.force_consistent)
             self.atoms.get_forces()
@@ -670,16 +664,16 @@ def make_neb(self, images_interpolation=None):
 @parallel_function
 def get_neb_predictions(images):
     neb_pred_energy = []
-    # neb_pred_forces = []
+    neb_pred_forces = []
     neb_pred_unc = []
+
     for i in images:
         neb_pred_energy.append(i.get_potential_energy())
-        # neb_pred_forces.append(i.get_forces())
+        neb_pred_forces.append(i.get_forces())
         unc = i.calc.results['uncertainty']
         neb_pred_unc.append(unc)
-    neb_pred_unc[0] = 0.0
-    neb_pred_unc[-1] = 0.0
-    predictions = {'energy': neb_pred_energy, 'uncertainty': neb_pred_unc}
+
+    predictions = {'energy': neb_pred_energy, 'forces': neb_pred_forces, 'uncertainty': neb_pred_unc}
 
     return predictions
 
