@@ -23,7 +23,8 @@ class ReferenceData(object):
                  data_type='float64',
                  data_process='batch',
                  soap_param=None,
-                 mask_constraints=False):
+                 mace_param=None,
+                 mask_constraints=True):
 
         self.data_process = data_process
         if data_type == 'float32':
@@ -35,10 +36,7 @@ class ReferenceData(object):
             self.numpy_data_type = np.float64
             self.torch_data_type = torch.float64
 
-        if device == 'cpu':
-            self.device = device
-        else:
-            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
+        self.device = device
 
         self.images = []
         self.lattice = np.array([], dtype=self.numpy_data_type)  # [Ndata, 3, 3] if pbc
@@ -75,6 +73,12 @@ class ReferenceData(object):
         else:
             self.soap_param = soap_param
 
+        if mace_param is None:
+            self.mace_param = {'system': "materials",
+                               'model': "small"}
+        else:
+            self.mace_param = mace_param
+
         self.mask_constraints = mask_constraints
         if self.mask_constraints:
             self.atoms_mask = self.create_mask()
@@ -254,20 +258,17 @@ class ReferenceData(object):
                 remove_indices.append(i)
 
         if remove_indices != []:
-            print("images", remove_indices, "are removed from training data since there are too close images")
-            print(", which can lead to ill-conditioned covariance")
+            print(f"Removed images {remove_indices} from training data (near-duplicates; risk of ill-conditioned covariance).")
+
             keep_images = [self.images[i] for i in keep_indices]
             self.read_structure_files(structure_files=keep_images, file_format='ase')
             self.set_data()
 
-    def config_calculator(self, kerneltype='sqexp', scale=0.4, weight=1.0, noise=1e-6, noisefactor=0.5,
+    def config_calculator(self, prior=None, prior_update=True, kerneltype='sqexp', scale=0.4, weight=1.0, noise=1e-6, noisefactor=0.5,
                           use_forces=True, sparse=None, sparse_derivative=None, autograd=False,
                           train_batch_size=25, eval_batch_size=25,
                           fit_weight=True, fit_scale=True):
 
-        # X_train_tensor = torch.as_tensor(self.fp, dtype=self.torch_data_type).to(self.device)
-        # dX_train_tensor = torch.as_tensor(self.dfp_dr, dtype=self.torch_data_type).to(self.device)
-
         if self.standardization:
             Y_train_tensor = torch.as_tensor(self.energy_scale, dtype=self.torch_data_type).to(self.device)
             dY_train_tensor = torch.as_tensor(self.force_scale, dtype=self.torch_data_type).to(self.device)
@@ -276,7 +277,9 @@ class ReferenceData(object):
             dY_train_tensor = torch.as_tensor(self.force, dtype=self.torch_data_type).to(self.device)
 
         if self.data_process == 'iterative':
-            self.calculator = gpr_iterative.GaussianProcess(kerneltype=kerneltype,
+            self.calculator = gpr_iterative.GaussianProcess(prior=prior,
+                                                            prior_update=prior_update,
+                                                            kerneltype=kerneltype,
                                                             scale=scale,
                                                             weight=weight,
                                                             noise=noise,
@@ -291,10 +294,13 @@ class ReferenceData(object):
                                                             data_type=self.data_type,
                                                             device=self.device,
                                                             soap_param=self.soap_param,
+                                                            mace_param=self.mace_param,
                                                             descriptor=self.descriptor,
                                                             atoms_mask=self.atoms_mask)
         elif self.data_process == 'batch':
-            self.calculator = gpr_batch.GaussianProcess(kerneltype=kerneltype,
+            self.calculator = gpr_batch.GaussianProcess(prior=prior,
+                                                        prior_update=prior_update,
+                                                        kerneltype=kerneltype,
                                                         scale=scale,
                                                         weight=weight,
                                                         noise=noise,
@@ -311,12 +317,13 @@ class ReferenceData(object):
                                                         data_type=self.data_type,
                                                         device=self.device,
                                                         soap_param=self.soap_param,
+                                                        mace_param=self.mace_param,
                                                         descriptor=self.descriptor,
                                                         atoms_mask=self.atoms_mask)
 
         self.calculator.train_model()
         if fit_scale:
-            self.fit_scale_only(candidates=5)
+            self.fit_scale_only()
         if fit_weight:
             self.fit_weight_only(use_forces=True)
 
@@ -327,20 +334,26 @@ class ReferenceData(object):
         k = weight**2 * sqexp(x,x',other hyperparameters)
         """
 
-        if use_forces:
-            _prior_array = self.calculator.prior.potential_batch(self.images, get_forces=True)
+        max_weight = 100.0
+        prev_weight = copy.deepcopy(self.calculator.weight)
+
+        try:
+            _prior_array = self.calculator.prior.potential_batch(len(self.images), len(self.images[0]))
             _factor = torch.sqrt(torch.matmul(self.calculator.YdY.flatten() - _prior_array,
                                               self.calculator.model_vector) / _prior_array.shape[0])
             self.calculator.weight *= _factor.squeeze()
             self.calculator.hyper_params.update(dict(weight=self.calculator.weight))
             self.calculator.kernel.set_params(self.calculator.hyper_params)
 
-        else:
-            _prior_array = self.calculator.prior.potential_batch(self.images, get_forces=False)
-            _n_train = len(self.images)
-            _factor = torch.sqrt(torch.matmul(self.calculator.YdY.flatten()[:_n_train] - _prior_array,
-                                              self.calculator.model_vector[:_n_train]) / _prior_array.shape[0])
-            self.calculator.weight *= _factor.squeeze()
+            if self.calculator.weight < max_weight:
+                pass
+            else:
+                raise ValueError(f"Weight parameter is too high ({self.calculator.weight}).")
+
+        except Exception as e:
+            print(f"{e}")
+            print("Fix the weight parameter")
+            self.calculator.weight = prev_weight
             self.calculator.hyper_params.update(dict(weight=self.calculator.weight))
             self.calculator.kernel.set_params(self.calculator.hyper_params)
 
@@ -350,7 +363,7 @@ class ReferenceData(object):
 
         return
 
-    def fit_scale_only(self, candidates=5, factor=4.0):
+    def fit_scale_only(self, candidates=3, factor=2.0):
         """
         Update the kernel scale keeping all other hyperparameters fixed by evaluating marginal likelihood over candidate scales.
 
@@ -368,7 +381,7 @@ class ReferenceData(object):
         best_logp = -torch.inf
 
         # Generate candidate scales logarithmically spaced around current scale
-        candidate_scales = current_scale * factor ** torch.linspace(-1, 1, candidates)
+        candidate_scales = current_scale * factor ** torch.linspace(-1, 1, candidates, device=self.device)
 
         print('Candidate scales:', candidate_scales)
         for candidate_scale in candidate_scales:
@@ -378,19 +391,31 @@ class ReferenceData(object):
             self.calculator.kernel.set_params(self.calculator.hyper_params)
 
             # Train GP with current candidate scale
-            self.calculator.train_model()
+            try:
+                self.calculator.train_model()
+            except Exception as e:
+                print(f"{e}")
+                continue
 
             # Evaluate marginal likelihood
             y_flat = self.calculator.YdY.flatten()
-            m_ = self.calculator.prior.potential_batch(self.images, get_forces=True)
+            m_ = self.calculator.prior.potential_batch(len(self.images), len(self.images[0]))
             a_ = self.calculator.model_vector
 
             logP = -0.5 * torch.matmul(y_flat - m_, a_) - torch.sum(torch.log(torch.diag(self.calculator.K_XX_L))) \
                    - self.calculator.Ntrain * 0.5 * torch.log(torch.as_tensor(2 * torch.pi, dtype=self.torch_data_type).to(self.device))
 
+            # MAP 보정: log-normal prior centered at previous (or initial) scale
+            log_l = torch.log(candidate_scale)
+            mu = torch.log(torch.tensor(0.4, dtype=self.torch_data_type, device=self.device))
+            sigma = torch.tensor(0.5, dtype=self.torch_data_type, device=self.device)
+            log_prior = -0.5 * ((log_l - mu) / sigma) ** 2 - log_l
+            lambda_prior = 1.0  # 튜닝
+            logP_total = logP + lambda_prior * log_prior
+
             # Keep track of best scale
-            if logP > best_logp:
-                best_logp = logP
+            if logP_total > best_logp:
+                best_logp = logP_total
                 best_scale = candidate_scale
 
         # Update to best scale found
@@ -406,19 +431,17 @@ class ReferenceData(object):
 
     def evaluation(self, get_variance=False):
 
-        self.calculator.eval()
-        with torch.no_grad():
-            if get_variance:
-                energy_gpr_scale, force_gpr_scale, uncertainty_gpr = self.calculator(eval_images=self.images,
-                                                                                     get_variance=get_variance)
+        if get_variance:
+            energy_gpr_scale, force_gpr_scale, uncertainty_gpr = self.calculator.eval_batch(eval_images=self.images,
+                                                                                            get_variance=get_variance)
 
-                return energy_gpr_scale.cpu().detach().numpy(), force_gpr_scale.cpu().detach().numpy(), uncertainty_gpr.cpu().detach().numpy()
+            return energy_gpr_scale.cpu().detach().numpy(), force_gpr_scale.cpu().detach().numpy(), uncertainty_gpr.cpu().detach().numpy()
 
-            else:
-                energy_gpr_scale, force_gpr_scale, _ = self.calculator(eval_images=self.images,
-                                                                       get_variance=get_variance)
+        else:
+            energy_gpr_scale, force_gpr_scale, _ = self.calculator.eval_batch(eval_images=self.images,
+                                                                              get_variance=get_variance)
 
-                return energy_gpr_scale.cpu().detach().numpy(), force_gpr_scale.cpu().detach().numpy(), None
+            return energy_gpr_scale.cpu().detach().numpy(), force_gpr_scale.cpu().detach().numpy(), None
 
     def standardize_energy_force(self, reference_training_energy):
         """
@@ -560,6 +583,7 @@ class ReferenceData(object):
             'species': self.species,
             'num_atom': self.num_atom,
             'soap_param': self.soap_param,
+            'mace_param': self.mace_param,
         }
         torch.save(state, file)
         self.calculator.save_data()
@@ -588,6 +612,7 @@ class ReferenceData(object):
         self.num_atom = state.get('num_atom')
 
         self.soap_param = state.get('soap_param')
+        self.mace_param = state.get('mace_param')
 
         if self.pbc:
             atom = Atoms(self.species)
@@ -609,6 +634,7 @@ class ReferenceData(object):
                                                             data_type=self.data_type,
                                                             device=self.device,
                                                             soap_param=self.soap_param,
+                                                            mace_param=self.mace_param,
                                                             descriptor=self.descriptor)
             self.calculator.load_data()
 
@@ -617,5 +643,6 @@ class ReferenceData(object):
                                                         data_type=self.data_type,
                                                         device=self.device,
                                                         soap_param=self.soap_param,
+                                                        mace_param=self.mace_param,
                                                         descriptor=self.descriptor)
             self.calculator.load_data()
